// translated implementation of the improved perlin function http://mrl.nyu.edu/~perlin/noise/
// COPYRIGHT 2002 KEN PERLIN

#include "PerlinNoise.h"
#include "random"
#include <algorithm>
#include <numeric>

PerlinNoise::~PerlinNoise()
{
}

// Initialize with the reference values for the permutation vector
PerlinNoise::PerlinNoise() {
	// Initialize the permutation vector with the reference values
	p = {
		151,160,137,91,90,15,131,13,201,95,96,53,194,233,7,225,140,36,103,30,69,142,
		8,99,37,240,21,10,23,190, 6,148,247,120,234,75,0,26,197,62,94,252,219,203,117,
		35,11,32,57,177,33,88,237,149,56,87,174,20,125,136,171,168, 68,175,74,165,71,
		134,139,48,27,166,77,146,158,231,83,111,229,122,60,211,133,230,220,105,92,41,
		55,46,245,40,244,102,143,54, 65,25,63,161,1,216,80,73,209,76,132,187,208, 89,
		18,169,200,196,135,130,116,188,159,86,164,100,109,198,173,186, 3,64,52,217,226,
		250,124,123,5,202,38,147,118,126,255,82,85,212,207,206,59,227,47,16,58,17,182,
		189,28,42,223,183,170,213,119,248,152, 2,44,154,163, 70,221,153,101,155,167,
		43,172,9,129,22,39,253, 19,98,108,110,79,113,224,232,178,185, 112,104,218,246,
		97,228,251,34,242,193,238,210,144,12,191,179,162,241, 81,51,145,235,249,14,239,
		107,49,192,214, 31,181,199,106,157,184, 84,204,176,115,121,50,45,127, 4,150,254,
		138,236,205,93,222,114,67,29,24,72,243,141,128,195,78,66,215,61,156,180 };
	// Duplicate the permutation vector
	p.insert(p.end(), p.begin(), p.end());
}

// Generate a new permutation vector based on the value of seed
PerlinNoise::PerlinNoise(unsigned int seed) {
	p.resize(256);
	
	// Fill p with values from 0 to 255
	std::iota(p.begin(), p.end(), 0);
	
	// Initialize a random engine with seed
	std::default_random_engine engine(seed);

	// Suffle  using the above random engine
	std::shuffle(p.begin(), p.end(), engine);

	// Duplicate the permutation vector
	p.insert(p.end(), p.begin(), p.end());
}

double PerlinNoise::Noise(double x)
{
	return Noise(x, 0.0, 0.0);
}

double PerlinNoise::Noise(double x, double y) {
	return Noise(x, y, 0.0);
}

double PerlinNoise::Noise(double x, double y, double z) {
	// Find the unit cube that contains the point
	int X = (int)floor(x) & 255;
	int Y = (int)floor(y) & 255;
	int Z = (int)floor(z) & 255;

	// Find relative x, y, z of point in cube
	x -= floor(x);
	y -= floor(y);
	z -= floor(z);

	// Compute Fade curves for each of x, y, z
	double u = Fade(x);
	double v = Fade(y);
	double w = Fade(z);

	// Hash coordinates of the 8 cube corners
	int A = p[X] + Y;
	int AA = p[A] + Z;
	int AB = p[A + 1] + Z;
	int B = p[X + 1] + Y;
	int BA = p[B] + Z;
	int BB = p[B + 1] + Z;

	// Add blended results from 8 corners of cube
	double res = Lerp(w,
		Lerp(v, 
			Lerp(u, 
				Grad(p[AA], x, y, z), 
				Grad(p[BA], x - 1, y, z)),
			Lerp(u, 
				Grad(p[AB], x, y - 1, z), 
				Grad(p[BB], x - 1, y - 1, z))),
		Lerp(v, 
			Lerp(u, 
				Grad(p[AA + 1], x, y, z - 1), 
				Grad(p[BA + 1], x - 1, y, z - 1)),
			Lerp(u, 
				Grad(p[AB + 1], x, y - 1, z - 1), 
				Grad(p[BB + 1], x - 1, y - 1, z - 1))));
	//maps values to [0,1]
	return (res + 1) / 2;
}

double PerlinNoise::OctaveNoise(double x, int octaves, double persistence) {
	double total = 0.0;
	double frequency = 1.0;
	double amplitude = 1.0;
	double maxValue = 0.0f;
	for (int i = 0; i<octaves; i++) {
		total += Noise(x * frequency) * amplitude;
		maxValue += amplitude;
		amplitude *= persistence;
		frequency /= persistence;
	}
	//maps values to [0,1] but doesnt grant to have value with 1
	return total / maxValue;
}

double PerlinNoise::OctaveNoise(double x, int octaves, double lacunarity, double changeInAmplitude)
{
	double total = 0.0;
	double frequency = 1.0;
	double amplitude = 1.0;
	double maxValue = 0.0f;
	for (int i = 0; i<octaves; i++) {
		total += Noise(x * frequency) * amplitude;
		maxValue += amplitude;
		amplitude *= lacunarity;
		frequency /= changeInAmplitude;
	}
	//maps values to [0,1] but doesnt grant to have value with 1
	return total / maxValue;
}

double PerlinNoise::OctaveNoise(double x, double y, int octaves, double persistence) {
	double total = 0.0;
	double frequency = 1.0;
	double amplitude = 1.0;
	double maxValue = 0.0f;
	for (int i = 0; i<octaves; i++) {
		total += Noise(x * frequency, y * frequency) * amplitude;
		maxValue += amplitude;
		amplitude *= persistence;
		frequency /= persistence;
	}
	//maps values to [0,1] but doesnt grant to have value with 1
	return total / maxValue;
}

double PerlinNoise::OctaveNoise(double x, double y, int octaves, double lacunarity, double changeInAmplitude)
{
	double total = 0.0;
	double frequency = 1.0;
	double amplitude = 1.0;
	double maxValue = 0.0f;
	for (int i = 0; i<octaves; i++) {
		total += Noise(x * frequency, y * frequency) * amplitude;
		maxValue += amplitude;
		amplitude *= lacunarity;
		frequency /= changeInAmplitude;
	}
	//maps values to [0,1] but doesnt grant to have value with 1
	return total / maxValue;
}

double PerlinNoise::OctaveNoise(double x, double y, double z, int octaves, double persistence) {
	double total = 0.0;
	double frequency = 1.0;
	double amplitude = 1.0;
	double maxValue = 0.0f;
	for (int i = 0; i<octaves; i++) {
		total += Noise(x * frequency, y * frequency, z * frequency) * amplitude;
		maxValue += amplitude;
		amplitude *= persistence;
		frequency /= persistence;
	}
	//maps values to [0,1] but doesnt grant to have value with 1
	return total / maxValue;
}

double PerlinNoise::OctaveNoise(double x, double y, double z, int octaves, double lacunarity, double changeInAmplitude)
{
	double total = 0.0;
	double frequency = 1.0;
	double amplitude = 1.0;
	double maxValue = 0.0f;
	for (int i = 0; i<octaves; i++) {
		total += Noise(x * frequency, y * frequency, z * frequency) * amplitude;
		maxValue += amplitude;
		amplitude *= lacunarity;
		frequency /= changeInAmplitude;
	}
	//maps values to [0,1] but doesnt grant to have value with 1
	return total;
}

float PerlinNoise::Fade(float t) {
	return t * t * t * (t * (t * 6.0f - 15.0f) + 10.0f);
}

float PerlinNoise::Lerp(float t, float a, float b) {
	return a + t * (b - a);
}

float PerlinNoise::Grad(int hash, float x, float y, float z) {
	int h = hash & 15;
	// Convert lower 4 bits of hash into 12 Gradient directions
	float u = h < 8 ? x : y,
		v = h < 4 ? y : h == 12 || h == 14 ? x : z;
	return ((h & 1) == 0 ? u : -u) + ((h & 2) == 0 ? v : -v);
	//switch (hash & 15) {
	//case 0: return x + y; // (1,1,0)
	//case 1: return -x + y; // (-1,1,0)
	//case 2: return x - y; // (1,-1,0)
	//case 3: return -x - y; // (-1,-1,0)
	//case 4: return x + z; // (1,0,1)
	//case 5: return -x + z; // (-1,0,1)
	//case 6: return x - z; // (1,0,-1)
	//case 7: return -x - z; // (-1,0,-1)
	//case 8: return y + z; // (0,1,1),
	//case 9: return -y + z; // (0,-1,1),
	//case 10: return y - z; // (0,1,-1),
	//case 11: return -y - z; // (0,-1,-1)
	//case 12: return y + x; // (1,1,0)
	//case 13: return -x + y; // (-1,1,0)
	//case 14: return -y + z; // (0,-1,1)
	//case 15: return -y - z; // (0,-1,-1)
	//}
}