#include "Segment.h"
#include <algorithm>
//#include "Building.h"

Segment::Segment()
{
}

// Sets default values
Segment::Segment(FVector2D _start, FVector2D _end, bool _isHighway)
{
	start = _start;
	end = _end;
	isHighway = _isHighway;
	direction = (end - start);
	length = direction.Size();
	originalLength = length;
	direction.Normalize();
	originalDirection = direction;
	isEndConnected = false;
	category = 1;
}

Segment::Segment(FVector2D _start, FVector2D _end, bool _isHighway, int _category)
{
	start = _start;
	end = _end;
	isHighway = _isHighway;
	direction = (end - start);
	length = direction.Size();
	originalLength = length;
	direction.Normalize();
	originalDirection = direction;
	isEndConnected = false;
	category = _category;
}

void Segment::SetStart(FVector2D newStart)
{
	start = newStart;
	direction = (GetEnd() - GetStart());
	length = direction.Size();
	direction.Normalize();
}

void Segment::SetEnd(FVector2D newEnd)
{
	end = newEnd;
	direction = (GetEnd() - GetStart());
	length = direction.Size();
	direction.Normalize();
}

FVector2D Segment::GetStart()
{
	return start;
}

FVector2D Segment::GetEnd()
{
	return end;
}

float Segment::GetOriginalLength()
{
	return originalLength;
}

FVector2D Segment::GetOriginalDirection()
{
	return originalDirection;
}

void Segment::Split(FVector2D splitLocation, QuadTree<Segment> qTree) {
	Segment splitPart = Segment(start, splitLocation, isHighway);
	qTree.AddSegment(splitPart);
	start = splitLocation;
}

bool Segment::OnSegment(FVector2D p, FVector2D q, FVector2D r)
{
	if (q.X <= std::max(p.X, r.X) && q.X >= std::min(p.X, r.X) &&
		q.Y <= std::max(p.Y, r.Y) && q.Y >= std::min(p.Y, r.Y))
		return true;

	return false;
}

int Segment::Orientation(FVector2D p, FVector2D q, FVector2D r)
{
	// https://www.geeksforgeeks.org/orientation-3-ordered-points/ 
	float val = (q.Y - p.Y) * (r.X - q.X) -
		(q.X - p.X) * (r.Y - q.Y);

	if (val == 0) return 0;  // colinear 

	return (val > 0) ? 1 : 2; // clock or counterclock wise 
}

bool Segment::CheckForIntersection(Segment segment)
{
	//if (start == segment.end || start == segment.start) return false;//dont report intersection if its the following object or if multiple streets start at the split

	// Find the four orientations
	int o1 = Orientation(start, end, segment.start);
	int o2 = Orientation(start, end, segment.end);
	int o3 = Orientation(segment.start, segment.end, start);
	int o4 = Orientation(segment.start, segment.end, end);

	// General case 
	if (o1 != o2 && o3 != o4)
		return true;

	// Special Cases 
	// p1, q1 and p2 are colinear and p2 lies on segment p1q1 
	if (o1 == 0 && OnSegment(start, segment.start, end)) return true;

	// p1, q1 and q2 are colinear and q2 lies on segment p1q1 
	if (o2 == 0 && OnSegment(start, segment.end, end)) return true;

	// p2, q2 and p1 are colinear and p1 lies on segment p2q2 
	if (o3 == 0 && OnSegment(segment.start, start, segment.end)) return true;

	// p2, q2 and q1 are colinear and q1 lies on segment p2q2 
	if (o4 == 0 && OnSegment(segment.start, end, segment.end)) return true;

	return false;
}

FVector2D Segment::GetIntersectionLocation(Segment segment) {
	float s1_x = end.X - start.X;
	float s1_y = end.Y - start.Y;
	float s2_x = segment.end.X - segment.start.X;
	float s2_y = segment.end.Y - segment.start.Y;

	float a = s2_x * (start.Y - segment.start.Y) - s2_y * (start.X - segment.start.X);
	float b = -s2_x * s1_y + s1_x * s2_y;
	
	float t = (s2_x * (start.Y - segment.start.Y) - s2_y * (start.X - segment.start.X)) / (-s2_x * s1_y + s1_x * s2_y);

	float x = start.X + (t * s1_x);
	float y = start.Y + (t * s1_y);
	
	return FVector2D(x, y);
}

int Segment::CheckStreetEndInside(Building building)
{
	Segment seg1 = Segment(start, FVector2D(999999, start.Y), false);
	Segment seg2 = Segment(end, FVector2D(999999, end.Y), false);
	int count1 = 0;
	int count2 = 0;
	for (int i = 0; i < building.layout.Num(); i++)
	{
		if (seg1.CheckForIntersection(Segment(building.layout[i], building.layout[(i + 1) % building.layout.Num()], false))) count1++;
		if (seg2.CheckForIntersection(Segment(building.layout[i], building.layout[(i + 1) % building.layout.Num()], false))) count2++;
	}
	//if odd number of intersections with polygon, then the point is inside
	if (count1 % 2 != 0 && count2 % 2 != 0) return 0; //segment completly inside
	else if (count1 % 2 != 0) return 1; //segment start inside
	else if (count2 % 2 != 0) return 2; //segment end inside
	else return 3;
}

Building Segment::CheckIntersectionBuilding(Building &building, bool createNewBuilding, float minSize, float width)
{
	int mod = building.layout.Num();
	int firstIntersectionIndex = -1;
	int secondIntersectionIndex = -1;
	FVector2D firstIntersectionLocation;
	FVector2D secondIntersectionLocation;
	Segment seg;
	switch (CheckStreetEndInside(building))
	{
	case 0:
		seg = Segment(start - 1000 * direction, end + 1000 * direction, false);
		break;
	case 1:
		seg = Segment(start - 1000 * direction, end, false);
		break;
	case 2:
		seg = Segment(start, end + 1000 * direction, false);
		break;
	default:
		seg = Segment(start, end, false);
		break;
	}

	for (int i = 0; i < building.layout.Num(); i++)
	{
		//to esure the index is below the intersection
		if (secondIntersectionIndex != -1) break;
		Segment tempSegment = Segment(building.layout[i], building.layout[(i + 1) % mod], false);
		FVector2D tempLocation = seg.GetIntersectionLocation(tempSegment);
		if (seg.CheckForIntersection(tempSegment) && !isnan(tempLocation.X)) {
			for (int l = i + 1; l < building.layout.Num(); l++)
			{
				Segment tempSegment2 = Segment(building.layout[l], building.layout[(l + 1) % mod], false);
				if (seg.CheckForIntersection(tempSegment2) && secondIntersectionIndex == -1) {
					secondIntersectionIndex = l;
				}
			}
			firstIntersectionIndex = i;
		}
	}
	
	if (secondIntersectionIndex == -1) return Building();
	else {
		Segment firstSegment = Segment(building.layout[firstIntersectionIndex], building.layout[firstIntersectionIndex + 1], false);
		Segment secondSegment = Segment(building.layout[secondIntersectionIndex], building.layout[(secondIntersectionIndex + 1) % mod], false);
		firstIntersectionLocation = GetIntersectionLocation(firstSegment) + secondSegment.direction;
		secondIntersectionLocation = GetIntersectionLocation(secondSegment) + secondSegment.direction;
		FVector2D offset = secondSegment.direction;
		offset *= width;
		Building newBuilding = building.Split(firstIntersectionLocation, secondIntersectionLocation, firstIntersectionIndex, secondIntersectionIndex, createNewBuilding, minSize, offset);
		if (newBuilding.layout.Num() > 0) return newBuilding;
		else return Building();
	}
}

float Segment::GetLowY() {
	return fminf(start.Y, end.Y);
}

float Segment::GetLowX() {
	return fminf(start.X, end.X);
}

float Segment::GetHighY() {
	return fmaxf(start.Y, end.Y);
}

float Segment::GetHighX() {
	return fmaxf(start.X, end.X);
}